#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/stat.h>
#include <ctype.h>
#include <string.h>

#define VERSION "amsolver.2006.03.18" 

struct Atom {
    int *metalist; // for a metaatom: list of prop atoms
                   // for a prop atom: list of metaatoms it appears in
    int *weightlist; // for a weight atom: list of weights corresponding to
                     // atoms from metalist
    int neg; // the sum of weights of prop atoms that must be false
    int pos; // the sum of weights of prop atoms that must be true
    int cnt; // the number of elements in metalist; for horn atoms, the number
             // of rules having it in the head
    int wgt; // the sum of weights of prop atoms in the metalist;
             // for cardinality atoms the same as cnt; for prop atoms undefined
    short horn; // 1 for horn atoms, 0 for non-horn
    short meta; // 0 for prop atoms, 1 for card atoms, 2 for weight atoms
    int hdcnt; // for a prop atom: how many times it appears in head
    int bdycnt; // for a prop atom: how many times it appears in body
    int *hdlst; // for a prop atom: list of rules containing it in the head
    int *bdylst; // for a prop atom: list of rules containing it in the body
};

struct AtomStk {
    int neg;
    int pos;
    int cnt;
    int wgt;
    short val;
    short req;
};

struct Rule {
    short horn;
    int hdcnt; // the number of atoms in the head
    int bdycnt; // the number of atoms in the body
    int *bdy; // list of atoms in the body
    int *hd; // list of atoms in the head
};

struct RuleStk {
    int sat; // 1 for satisfied rules, 0 for unsatisfied
    int hdcnt; // the number of unassigned atoms in the head
    int bdycnt; // the number of unassigned atoms in the body
    double prec; // precedence of the rule
};

struct BackTrack {
    int atom;
    short val;
    int satisfied;
    struct RuleStk *rlst;
    struct AtomStk *alst;
    struct BackTrack *up;
    struct BackTrack *down;
};

struct BackTrack *BTtop=NULL;

int LAH=0;
int Backtrack=0;
int Answer=0;
int Nvar=0; // the total number of prop and metaatoms
int Outvar=0; // the number of prop atoms
int Nrule=0; // the number of rules in the theory
int Nsat=0; // the number of non-horn rules
int Sat=0; // the number of satisfied non-horn rules
int Count=0; // the number of solutions required
int Cnt=0; // the number of solutions found
int Maxlit=0; // the largest number of atoms in a rule or a metaatom
int SetAtomCnt=0; 
int Mult=13; // arbitrary value
int Meta=0; // used to track numbers assigned to metaatoms when the theory is read

struct Atom *Atoms; //list of atoms
struct Rule *Rules; //list of rules
struct RuleStk *RLst; 
struct AtomStk *Alst;
char **At; // names of atoms

double *Prec; // precedence of rules
double *Tmplist; // precedence of atoms, used when an atom is selected for branching
int *SetAtomQue;
short *AtomQueVal;
short *PathList; // used to verify whether a horn atom can be proved by horn rules

int Print=0;
int Branch=0;
FILE *fptr;
int ShowCnt=0;
char **ShowAtoms;

// creates a new BackTrack element and initializes all components to 0
struct BackTrack *newStck()
{
    struct BackTrack *tmp;

    if ((tmp=(struct BackTrack *)malloc(sizeof(struct BackTrack)))==NULL) {
        fprintf(stderr, "out of memory\n");
        exit(1);
    }
    memset(tmp, 0, sizeof(struct BackTrack));
    return tmp;
}

int *newData(size_t number, int val)
{
    int *tmp;
      
    if (!number)
        return NULL;
    if ((tmp=(int *)malloc(number))==NULL) {
        fprintf(stderr, "cannot allocate Clause list\n");
        exit(1);
    }
    tmp=(int *)memset(tmp, val, number);
    return tmp;
}

// reads the index of the atom and reads its name into At[index]
void parseTxtVar(char *input)
{
    int i=0, len, index;
    char buffer[1000000];

    len=strlen(input)-1;
    while (i<len && !isdigit((int)input[i]))
        i++;
    index=atoi(&input[i]);
    while (i<len && isdigit((int)input[i]))
        i++;
    input[len]='\0';
    while (i<len && isspace((int)input[i]))
        i++;
    if (index>0)
        At[index]=strdup(&input[i]);
    else {
        if (index==-1)
            index=Nvar+1;
        sprintf(buffer, "%s\n%s", At[index], &input[i]); 
        free(At[index]);
        At[index]=strdup(buffer);
    }
}

// returns first integer from the input and updates begin
int parseInt(char *input, int *begin)
{
    int i=*begin, j=0;
    char line[180];

    while (input[i]!='\n' && isspace((int)input[i]))
        i++;
    if (input[i]=='\n' || input[i]=='}' || input[i]==']')
        return -1;
    while (isdigit((int)input[i]))
        line[j++]=input[i++];
    line[j]='\n';
    (*begin)=i;
    return atoi(line);
}

// reads list of prop atoms for a metaatom and returns it,
// updates index and count,
// memory for the list is allocated dynamically
int *getList(char *input, int *index, int *count)
{
    int i=(*index), j, size=0;
    int *tmp=NULL;

    while ((parseInt(input, &i)>=0))
        size++;
    if (size) {
        tmp=(int *)newData(sizeof(int)*size, 0);
        (*count)=size;
        i=(*index);
        for (j=0; j<size; j++)
            tmp[j]=parseInt(input, &i);
    }
    (*index)=i;
    return tmp;
}

// reads a metaatom
int parseAtom(char *input, int *index) 
{
    int count=0, pos, neg, *list;
    char buffer[180];

    (*index)++;
    pos=parseInt(input, index);
    neg=parseInt(input, index);
    list=getList(input, index, &count);
    if (Maxlit<count)
        Maxlit=count;
    Meta++;
    sprintf(buffer, "Meta%i", Meta);
    At[Meta]=strdup(buffer);
    Atoms[Meta].meta=1;
    Atoms[Meta].pos=pos;
    Atoms[Meta].neg=(count-neg<0) ? 0 : count-neg; // so many atoms must be false
    Atoms[Meta].metalist=list;
    Atoms[Meta].wgt=Atoms[Meta].cnt=count;
    return Meta;
}

int parseEqual(char *input, int *i)
{
    while (input[*i]!='\n' && isspace((int)input[*i]))
        (*i)++;
    if (input[*i]=='=') {
        (*i)++;
        return 1;
    } else
        return 0;
}

int parseW(char *input, int *index)
{
    if (parseInt(input, index)>=0) 
        if (parseEqual(input, index))
            return parseInt(input, index);
    return -1;
}

// reads list of prop atoms for a weight atom and returns it,
// updates index, count, weights,
// memory for the list and weights is allocated dynamically
int *getWList(char *input, int *index, int *count, int **weight)
{
    int i=(*index), j, size=0;
    int *tmp=NULL;

    while ((parseW(input, &i)>=0))
        size++;
    if (size) {
        tmp=(int *)newData(sizeof(int)*size, 0);
        *weight=(int *)newData(sizeof(int)*size, 0);
        (*count)=size;
        i=(*index);
        for (j=0; j<size; j++) {
            tmp[j]=parseInt(input, &i);
            parseEqual(input, &i);
            (*weight)[j]=parseInt(input, &i);
        }
    }
    (*index)=i;
    return tmp;
}

// sorts list of atoms in a weight atom in descending order of weights
void sortWList(int **list, int **weights, int count)
{
    int i, j, temp, sorted=0;

    // bubble sort
    for (i=1; i<count && !sorted; i++)
        for (sorted=1, j=0; j<count-i; j++)
            if ((*weights)[j] < (*weights)[j+1]) {
                temp=(*weights)[j];
                (*weights)[j]=(*weights)[j+1];
                (*weights)[j+1]=temp;
                temp=(*list)[j];
                (*list)[j]=(*list)[j+1];
                (*list)[j+1]=temp;
                sorted=0;
            }
}

// reads a weight atom
int parseWAtom(char *input, int *index) 
{
    int count=0, pos, neg, *list, *weights, sumw, i;
    char buffer[180];

    (*index)++;
    pos=parseInt(input, index);
    neg=parseInt(input, index);
    list=getWList(input, index, &count, &weights);
    if (Maxlit<count)
        Maxlit=count;
    Meta++;
    sprintf(buffer, "Meta%i", Meta);
    At[Meta]=strdup(buffer);
    Atoms[Meta].meta=2;
    Atoms[Meta].pos=pos;
    for (sumw=i=0; i<count; i++)
        sumw += *(weights+i);
    Atoms[Meta].neg=(sumw-neg<0) ? 0 : sumw-neg; // weights of atoms to be false
    sortWList(&list, &weights, count);
    Atoms[Meta].metalist=list;
    Atoms[Meta].weightlist=weights;
    Atoms[Meta].cnt=count;
    Atoms[Meta].wgt=sumw;
    return Meta;
}

// reads a rule
// invoked by getProblem()
int parseRule(char *input, int index)
{
    int i=0, count=0, hd=0;
    
    // count atoms in the body and in the head
    while (i<strlen(input)) {
        if (input[i]=='{' || input[i]=='[') {
            if (hd)
                Rules[index].hdcnt++;
            else
                Rules[index].bdycnt++;
            while (input[i]!='}' && input[i]!=']')
                i++;
        } else if (input[i]==',')
            hd=1;
        else if (input[i]==':') {
            hd=1;
            Rules[index].horn=1;
        } else if (isdigit((int)input[i])) {
            if (hd)
                Rules[index].hdcnt++;
            else
                Rules[index].bdycnt++;
            while (isdigit((int)input[i]))
                i++;
        }
        i++;
    }
    if (Rules[index].hdcnt+Rules[index].bdycnt==0)
        return 0;
    // read atoms for the body and for the head
    Rules[index].hd=(int *)newData(sizeof(int)*Rules[index].hdcnt, 0);
    Rules[index].bdy=(int *)newData(sizeof(int)*Rules[index].bdycnt, 0);
    Rules[index].hdcnt=Rules[index].bdycnt=i=hd=0;
    while (i<strlen(input)) {
        if (input[i]==',' || input[i]==':')
            hd=1;
        else if (input[i]=='{') {
            if (hd)
                Rules[index].hd[Rules[index].hdcnt++]=parseAtom(input, &i);
            else
                Rules[index].bdy[Rules[index].bdycnt++]=parseAtom(input, &i);
        } else if (input[i]=='[') {
            if (hd)
                Rules[index].hd[Rules[index].hdcnt++]=parseWAtom(input, &i);
            else
                Rules[index].bdy[Rules[index].bdycnt++]=parseWAtom(input, &i);
        } else if (isdigit((int)input[i])) {
            if (hd)
                Rules[index].hd[Rules[index].hdcnt++]=parseInt(input, &i);
            else
                Rules[index].bdy[Rules[index].bdycnt++]=parseInt(input, &i);
        }
        i++;
    }
    if (!Rules[index].horn)
        Nsat++;
    count=Rules[index].hdcnt+Rules[index].bdycnt;
    if (Maxlit<count)
        Maxlit=count;
    return 1;
}

// invoked by logic()
void setCnt() 
{
    int i, increase=1;

    for (i=1; i<=Outvar; i++)
        if (Alst[i].val==-1) // the value of prop atom i does not matter
            increase*=2; // thus, there are two solutions (i can be true or false)
    Cnt+=increase;
}

// invoked by logic() and PreProcess()
void printAnswer()
{
    int i, j, k;
    char buffer[180];

    if (Cnt==0) {
        fprintf(stdout, "\nAtoms determined during grounding:%s\n", At[0]);
        fprintf(stdout, "\nAtoms irrelevant:%s\n", At[Nvar+1]);
    }
    for (j=0; j<ShowCnt; j++)
        printf(" %s\n", ShowAtoms[j]);
    fprintf(stdout, "\nAnswer-set %i:\n", Cnt+1);
    for (i=1; i<=Outvar; i++)
        if (Alst[i].val) {
            if (!ShowCnt) {
                if (Alst[i].val==-1)
                    printf(" not %s or", At[i]);
                printf(" %s\n", At[i]);
            } else {
                for (j=0; j<strlen(At[i]) && (isdigit(At[i][j]) || isspace(At[i][j])); j++)
                    ;
                k=0;
                buffer[k]='\0';
                while (j<strlen(At[i]) && At[i][j]!='(')
                    buffer[k++]=At[i][j++];
                buffer[k]='\0';
                for (j=0; j<ShowCnt; j++) 
                    if (!strcmp(buffer, ShowAtoms[j])) {
                        if (Alst[i].val!=1)
                            printf(" not %s or", At[i]);
                        printf(" %s\n", At[i]);
                    }
            }
        }
    fprintf(stdout, "\n");
}

// invoked by setUp(), reads the input theory
int getProblem()
{
    int index=0, i;
    char line[1800];

    Nsat=0;
    // skip initial lines not starting with a 'p' 
    while (fgets(line, 1799, fptr) && line[0]!='p')
            ;
    if (line[0]=='p') {
        i=1; 
        Outvar=Meta=parseInt(line, &i);
        i++;
        Nvar= Outvar + parseInt(line, &i);
        i++;
        Nrule=parseInt(line, &i);
    } else {
        printf("error: in format of line 1 of input \n");
        return -1;
    }
    Atoms = (struct Atom *)newData(sizeof(struct Atom)*(Nvar+1), 0);
    At = (char **)newData(sizeof(char *)*(Nvar+2), 0); 
    At[0] = strdup("");
    At[Nvar+1] = strdup("");
    Rules=(struct Rule *)newData(sizeof(struct Rule)*Nrule, 0);
    // read rules and descriptions of atoms
    while (fgets(line, 1799, fptr)) {
        switch(line[0]) {
        case 'p':
            break;
        case 'c':
            parseTxtVar(line);
            break;
        default: 
            if (!parseRule(line, index++))
                return 0; 
            break;
        }
    }
    if (fptr!=stdin)
        fclose(fptr);
    if (Nrule!=index || Meta!=Nvar) {
        printf("Input Error: Nrule=%i index=%i - Meta=%i Nvar=%i\n", Nrule, index, Meta, Nvar);
        return -1;
    }
    SetAtomQue=newData(sizeof(int)*Nvar, 0);
    AtomQueVal=(short *)newData(sizeof(short)*(Nvar+1), -1);
    PathList=(short *)newData(sizeof(short)*(Nvar+1), 0);
    Tmplist=(double *)newData(sizeof(double)*(Nvar+1), 0);
    return 1;
}

// invoked by pushStck()
// creates and returns a new copy of Rlist
struct RuleStk *setCurrentRules()
{
    struct RuleStk *tmp;
    
    if ((tmp=(struct RuleStk *)malloc(sizeof(struct RuleStk)*Nrule))==NULL) {
        fprintf(stderr, "out of memory\n");
        exit(1);
    }
    memcpy(tmp, RLst, sizeof(struct RuleStk)*Nrule);
    return tmp;
}

// invoked by pushStck()
// creates and returns a new copy of Alist
struct AtomStk *setCurrentAtoms()
{
    struct AtomStk *tmp;

    if ((tmp=(struct AtomStk *)malloc(sizeof(struct AtomStk)*(Nvar+1)))==NULL) {
        fprintf(stderr, "out of memory\n");
        exit(1);
    }
    memcpy(tmp, Alst, sizeof(struct AtomStk)*(Nvar+1));
    return tmp;
}

// invoked by setUp()
// creates and initializes Rlist and Alist
void initializeCurrent()
{
    int i;

    if ((Alst=(struct AtomStk *)malloc(sizeof(struct AtomStk)*(Nvar+1)))==NULL) {
        fprintf(stderr, "out of memory\n");
        exit(1);
    }
    if ((RLst=(struct RuleStk *)malloc(sizeof(struct RuleStk)*Nrule))==NULL) {
        fprintf(stderr, "out of memory\n");
        exit(1);
    }
    memset(RLst, 0, sizeof(struct RuleStk)*Nrule);
    for (i=0; i<Nrule; i++) {
        RLst[i].bdycnt=Rules[i].bdycnt;
        RLst[i].hdcnt=Rules[i].hdcnt;
    }
    memset(Alst, -1, sizeof(struct AtomStk)*(Nvar+1));
    for (i=1; i<=Nvar; i++) {
        Alst[i].pos=Atoms[i].pos;
        Alst[i].neg=Atoms[i].neg;
        Alst[i].cnt=Atoms[i].cnt;
        Alst[i].wgt=Atoms[i].wgt;
    }
}    

// invoked by logic() and lookAHead()
void pushStck(int lit, int val)
{
    if (!BTtop)
        BTtop=newStck();
    else if (!BTtop->up) {
        BTtop->up=newStck();
        BTtop->up->down=BTtop;
        BTtop=BTtop->up;
    } else
        BTtop=BTtop->up;
    BTtop->alst=setCurrentAtoms();
    BTtop->rlst=setCurrentRules();
    BTtop->atom=lit;
    BTtop->val=val;
    BTtop->satisfied=Sat;
}

// invoked by logic() and lookAHead()
// when the stack is empty, 0 is returned and *val is undefined
// otherwise the literal from the top of stack is returned
int popStck(int *val)
{
    int lit=0;

    if (!BTtop)
        return 0;
    free(RLst);
    RLst=BTtop->rlst;
    Sat=BTtop->satisfied;
    lit=BTtop->atom;
    (*val)=BTtop->val;
    free(Alst);
    Alst=BTtop->alst;
    BTtop=BTtop->down;
    return lit;
}

// invoked by setUp()
// creates lists of rules containing each of atoms
void setReference()
{
    int i, j, aptr;

    for (j=0; j<Nrule; j++) {
        // for every atom count how many times this atom appears in body of a rule
        for (i=0; i<Rules[j].bdycnt; i++)
            Atoms[Rules[j].bdy[i]].bdycnt++;
        // for every atom count how many times this atom appears in head of a rule
        for (i=0; i<Rules[j].hdcnt; i++) {
            Atoms[Rules[j].hd[i]].hdcnt++;
            // mark atoms in heads of horn rules as horn atoms
            if (Rules[j].horn)
                Atoms[Rules[j].hd[i]].horn=1;
        }
    }
    // for every atom count how many times this atom appears in a metalist
    for (i=1; i<=Nvar; i++)
        if (Atoms[i].meta)
            for (j=0; j<Atoms[i].cnt; j++)
                Atoms[Atoms[i].metalist[j]].cnt++;
    // for every prop atom create a list of metaatoms containing this atom
    for (i=1; i<=Nvar; i++) {
        Atoms[i].bdylst=newData(sizeof(int)*(Atoms[i].bdycnt), 0);
        Atoms[i].hdlst=newData(sizeof(int)*(Atoms[i].hdcnt), 0);
        if (!Atoms[i].meta) {
            Atoms[i].metalist=newData(sizeof(int)*(Atoms[i].cnt), 0);
            Atoms[i].cnt=0;
        }
        Atoms[i].bdycnt=0;
        Atoms[i].hdcnt=0;
    }
    // for every prop atom create a list of rules containing this atom
    for (j=0; j<Nrule; j++) {
        for (i=0; i<Rules[j].bdycnt; i++)
            Atoms[Rules[j].bdy[i]].bdylst[Atoms[Rules[j].bdy[i]].bdycnt++]=j;
        for (i=0; i<Rules[j].hdcnt; i++)
            Atoms[Rules[j].hd[i]].hdlst[Atoms[Rules[j].hd[i]].hdcnt++]=j;
        if (Rules[j].horn)
            Atoms[Rules[j].hd[0]].cnt++;
    }
    // for every prop atom, create a list of metaatoms this atom appears in
    for (i=1; i<=Nvar; i++)
        if (Atoms[i].meta)
            for (j=0; j<Atoms[i].cnt; j++) {
                aptr=Atoms[i].metalist[j];
                Atoms[aptr].metalist[Atoms[aptr].cnt++]=i;
            }
}

// invoked by setUp()
// writes theory to standard output
void printTheory()
{
    int i, j;

    for (i=0; i<Nrule; i++) {
        printf("\nRule %i  %i: body count %i head count %i -- bodylist:",
                i, Rules[i].horn, Rules[i].bdycnt, Rules[i].hdcnt);
        for (j=0; j<Rules[i].bdycnt; j++)
            printf(" %i ", Rules[i].bdy[j]);
        printf(" -- headlist:");
        for (j=0; j<Rules[i].hdcnt; j++)
            printf(" %i ", Rules[i].hd[j]);
    }
    for (i=1; i<=Nvar; i++) {
        printf("\nAtom %i %s value = %i -- ", i, At[i], Alst[i].val);
            printf("Atomlist:");
        if (Atoms[i].meta)
            printf("MetaList:");
        for (j=0; j<Atoms[i].cnt; j++) {
            printf(" %i ", Atoms[i].metalist[j]);
            if (Atoms[i].meta==2)
                printf(" =%i ", Atoms[i].weightlist[j]);
        }
    }
    printf("\n");
}

void removeRef(int rule)
{
    if (!RLst[rule].sat) { // rule not satisfied so far
        if (!Rules[rule].horn)
            Sat++;
        else
            Alst[Rules[rule].hd[0]].cnt--;
    }
    RLst[rule].sat=1; // mark it satisfied
}

// invoked by PropAtoms()
int setMeta(int ltl, int val)
{
    int setval=-1, i, at, pos, neg, wgt;
    int meta=Atoms[ltl].meta;
     
    if (Alst[ltl].req>=0 && Alst[ltl].req!=val)
        return 0;
    Alst[ltl].req=val;
    pos=Alst[ltl].pos;
    neg=Alst[ltl].neg;
    wgt=Alst[ltl].wgt;
    if (val) { // metaatom must be true
        if (neg<=0 && pos<=0)
            return 1;
        else if (wgt<pos || wgt<neg || wgt<pos+neg) // it cannot be made true
            return 0;
        else if (neg<=0 && pos==wgt)
            setval=1; // make all unassigned atoms in the metalist true
        else if (pos<=0 && neg==wgt)
            setval=0; // make all unassigned atoms in the metalist false
    } else // metaatom must be false
    if (wgt<pos || wgt<neg || wgt<pos+neg) // it cannot be made true
        return 1;
    else if (meta==1 && Alst[ltl].cnt==1) { // only one unassigned atom in the metalist
        if (neg==1) 
            setval=1; // make it true to make the metaatom false
        else if (pos==1)
            setval=0; // make it false to make the metaatom false
    }
    if (setval>=0) // give all unassigned atoms in the metalist required value
        for (i=0; i<Atoms[ltl].cnt; i++) {
            at=Atoms[ltl].metalist[i];
            if (Alst[at].val==-1) {
                if (AtomQueVal[at]==-1) {
                    SetAtomQue[SetAtomCnt++]=at;
                    AtomQueVal[at]=setval;
                } else if (AtomQueVal[at]!=setval)
                    return 0;
            }
        }
    else
        return -1;
    return 1;
} 

// invoked by PropAtoms()
int setHorn(int ltl, int horn, int val)
{
    if (Alst[ltl].req>=0 && Alst[ltl].req!=val)
        return 0;
    Alst[ltl].req=val;
    if (!horn)
        return -1;
    return 1;
} 

// invoked by evaluateRule()
int propAtoms(int *list, int cnt, int horn, int val)
{
    int i, set=-1, ret;

    for (i=0; i<cnt; i++)
        if (Alst[list[i]].val<0) {
            if (AtomQueVal[list[i]]==-1) {
                ret=1;
                if (Atoms[list[i]].horn) {
                    if ((ret=setHorn(list[i], horn, val))==0)
                        return 0;
                } else if (Atoms[list[i]].meta) {
                    if ((ret=setMeta(list[i], val))==0)
                        return 0;
                }
                if (ret==1) {
                    SetAtomQue[SetAtomCnt++]=list[i];
                    AtomQueVal[list[i]]=val;
                } else if (ret==-1)
                    set=1;
            } else if (AtomQueVal[list[i]]!=val)
                return 0;
        }
    return set;
}

// invoked by CheckHorn(), recursive
int closure(int rule) // rule is a horn rule and its head must be true
{
    int i, j, at, r;

    for (i=0; i<Rules[rule].bdycnt; i++) {
        at=Rules[rule].bdy[i]; // at is in the body of rule
        if (Alst[at].val<0 && AtomQueVal[at]<0 && Atoms[at].horn) {
            if (PathList[at]) // circular dependency of horn atoms
                return 0;
            if (Alst[at].cnt==1) { // there is only one horn rule with the head at
                PathList[at]=1;
                for (j=0; j<Atoms[at].hdcnt; j++) { 
                    r=Atoms[at].hdlst[j]; // check all rules r with at in the head
                    if (Rules[r].horn && !RLst[r].sat && !closure(r))
                        return 0;
                }
            } else if (!Alst[at].cnt) // no horn rules with the head at
                return 0;
        }
    }
    for (i=0; i<Rules[rule].bdycnt; i++) {
        at=Rules[rule].bdy[i]; // at is in the body of rule
        if (Alst[at].val<0 && AtomQueVal[at]<0 && !Atoms[at].horn) {
            SetAtomQue[SetAtomCnt++]=at;
            AtomQueVal[at]=1;
        }
    }
    return 1;
}

//invoked by propRules() and preProcess()
int checkHorn(int at)
{
    int i, r;

    if (AtomQueVal[at]>=0)
        return 1;
    if (Alst[at].cnt>1)
        return 1;
    else if (!Alst[at].cnt) {
        if (Alst[at].req==1)
            return 0;
        Alst[at].req=0;
        SetAtomQue[SetAtomCnt++]=at;
        AtomQueVal[at]=0;
    } else if (Alst[at].cnt==1) {
        if (Alst[at].req==1) {
            memset(PathList, 0, sizeof(short)*(Nvar+1));
            PathList[at]=1;
            for (i=0; i<Atoms[at].hdcnt; i++) {
                r=Atoms[at].hdlst[i]; // at appears in the head of r
                if (Rules[r].horn && !RLst[r].sat && !closure(r))
                    return 0;
            }
        }
    }
    return 1;
}

int evaluateRule(int rule);

// invoked by Propagate()
int propRules(int at, int val)
{
    int i, j, setval, atmp, atemp, r, pos, neg, cnt, wgt, w, meta;

    if (!Atoms[at].meta && !Atoms[at].horn) {
        for (i=0; i<Atoms[at].cnt; i++) {
            atmp=Atoms[at].metalist[i]; // atmp is a metaatom at appears in
            if (Alst[atmp].val<0 && AtomQueVal[atmp]<0) {
                cnt=--(Alst[atmp].cnt);
                if ((meta=Atoms[atmp].meta)==2) { // weight atom
                    for (j=0; j<Atoms[atmp].cnt; j++)
                        if (Atoms[atmp].metalist[j]==at)
                            break;
                    w=Atoms[atmp].weightlist[j];
                    Alst[atmp].wgt -= w;
                    if (val) {
                        Alst[atmp].pos -= w;
                        if (Alst[atmp].pos<0)
                            Alst[atmp].pos=0; 
                    } else {
                        Alst[atmp].neg -= w;
                        if (Alst[atmp].neg<0)
                            Alst[atmp].neg=0; 
                    }
                } else { // cardinality atom
                    Alst[atmp].wgt--;
                    if (val)
                        Alst[atmp].pos--;
                    else
                        Alst[atmp].neg--;
                }
                pos=Alst[atmp].pos;
                neg=Alst[atmp].neg;
                wgt=Alst[atmp].wgt;
                if (wgt<pos || wgt<neg || wgt<pos+neg)
                    setval=0; // metaatom is false
                else if (pos<=0 && neg<=0)
                    setval=1; // metaatom is true
                else
                    setval=-1; // metaatom has unknown value
                if (setval>=0) {
                    if (Alst[atmp].req>=0 && Alst[atmp].req!=setval)
                        return 0;
                    SetAtomQue[SetAtomCnt++]=atmp;
                    AtomQueVal[atmp]=setval;
                } else if (Alst[atmp].req>=0) {
                    if (Alst[atmp].req) {
                        if (pos<=0 && neg==wgt)
                            setval=0; // make all atoms in the metalist false
                        else if (neg<=0 && pos==wgt)
                            setval=1; // make all atoms in the metalist true
                    } else if (meta==1 && cnt==1) {
                        if (neg==1)
                            setval=1;
                        else if (pos==1) 
                            setval=0;
                    }
                    if (setval>=0) {
                        for (j=0; j<Atoms[atmp].cnt; j++) {
                            atemp=Atoms[atmp].metalist[j];
                            if (Alst[atemp].val<0) {
                                if (AtomQueVal[atemp]==-1) {
                                    SetAtomQue[SetAtomCnt++]=atemp;
                                    AtomQueVal[atemp]=setval;
                                } else if (AtomQueVal[atemp]!=setval)
                                    return 0;
                            }
                        }
                        SetAtomQue[SetAtomCnt++]=atmp;
                        AtomQueVal[atmp]=Alst[atmp].req;
                    }
                }
            }
        }
    }
    for (i=0; i<Atoms[at].hdcnt; i++) {
        r=Atoms[at].hdlst[i]; // at appears in the head of r
        if (!RLst[r].sat) {
            if (val)
                removeRef(r);
            else {
                RLst[r].hdcnt--;
                if (!evaluateRule(r))
                    return 0;
                RLst[r].prec*=Mult;
            }
        }
    }
    for (i=0; i<Atoms[at].bdycnt; i++) {
        r=Atoms[at].bdylst[i]; // at appears in the body of r
        if (!RLst[r].sat) {
            if (val) {
                RLst[r].bdycnt--;
                if (!evaluateRule(r))
                    return 0;
                RLst[r].prec*=Mult;
            } else {
                removeRef(r);
                if (Rules[r].horn && !checkHorn(Rules[r].hd[0]))
                    return 0;
            }
        }
    }
    return 1;
}

// invoked by Propagate()
int setAtom(int ltl, int val)
{
    if (Alst[ltl].val<0) {
        Alst[ltl].val=val;
        return 1;
    } else
        return 0;
} 

// invoked by unitProp() and preProcess()
int propagate()
{
    int i;

    for (i=0; i<SetAtomCnt; i++) {
        if (!setAtom(SetAtomQue[i], AtomQueVal[SetAtomQue[i]]))
            return 0;
        if (!propRules(SetAtomQue[i], AtomQueVal[SetAtomQue[i]]))
            return 0;
    }
    if (Sat>=Nsat)
        return -1;
    return 1;
}

// invoked by logic() and lookAHead()
int unitProp(int literal, int val)
{
    SetAtomCnt=0;
    memset(AtomQueVal, -1, sizeof(short)*(Nvar+1));
    SetAtomQue[SetAtomCnt++]=literal;
    AtomQueVal[literal]=val;
    return propagate();
}

// invoked by logic() and lookAHead()
// returns the number of atoms copied to array literal
int getNextLit(int *literal)
{
    int i, j, k, cnt, r, mcnt=0, req=0, found=0, metamin=Nvar, min=Nvar;
    int lit=0; // an atom with the largest precedence max
    double posmax=0, max=0, metamax=0; 

    memset(Tmplist, 0, sizeof(double)*(Nvar+1));
    memset(literal, 0, sizeof(int)*(Nvar+1));
    for (i=1; i<=Outvar; i++) // for every prop atom i
        if (Alst[i].val<0) { // with no value
            if (!Atoms[i].horn) { // i is not a horn atom
                for (j=0; j<Atoms[i].hdcnt; j++)
                    if (!RLst[Atoms[i].hdlst[j]].sat) 
                        Tmplist[i] += RLst[Atoms[i].hdlst[j]].prec;
                for (j=0; j<Atoms[i].bdycnt; j++) {
                    r=Atoms[i].bdylst[j]; // atom i appears in the body of r
                    if (!RLst[r].sat) {
                        Tmplist[i] += RLst[r].prec;
                        if (Rules[r].horn && !found) {
                            for (cnt=0, k=0; k<Rules[r].bdycnt; k++)
                                if (Alst[Rules[r].bdy[k]].val>0 || !Atoms[Rules[r].bdy[k]].horn)
                                    cnt++;
                            if (cnt==Rules[r].bdycnt) // r is a horn rule such that every
                                found=1; // atom in the body either has a value or is not horn
                        }
                    }
                }
                for (j=0; j<Atoms[i].cnt; j++) // for metaatoms with no value containing i
                    if (Alst[Atoms[i].metalist[j]].val<0)
                        Tmplist[i] += Prec[Alst[Atoms[i].metalist[j]].cnt];
                if (Tmplist[i]>max) {
                    max=Tmplist[i]; 
                    lit=i;
                }
            } else if (Alst[i].req) // i is a horn atom
                req=1;
        }
    if (lit) { // there is an unassigned non-horn atom with a positive precedence
        for (i=Outvar+1; i<=Nvar; i++) { // for every metaatom i
            if (Alst[i].val<0) { // with no value
                if (req && !found)
                    for (j=0; j<Atoms[i].bdycnt; j++)
                        if (!RLst[Atoms[i].bdylst[j]].sat && Rules[Atoms[i].bdylst[j]].horn)
                            found=1; // found an unsat horn rule containing i
                if (Alst[i].req==1 && Alst[i].cnt<=metamin && Atoms[i].meta==1 &&
                        // in a cardinality atom with the shortest list, either 1 atom in the 
                        // list must be true and the others must be false or vice versa
                        ((Alst[i].pos==1 && 1+Alst[i].neg==Alst[i].cnt) 
                        || (Alst[i].neg==1 && 1+Alst[i].pos==Alst[i].cnt))) {
                    for (posmax=0, j=0; j<Atoms[i].cnt; j++) // get the total precedence
                        if (Alst[Atoms[i].metalist[j]].val<0) // of unassigned atoms
                            posmax += Tmplist[Atoms[i].metalist[j]]; // in the metalist of i
                    if (metamax<posmax || Alst[i].cnt<metamin || Atoms[i].cnt<min) { 
                        metamin=Alst[i].cnt; // i is a best metaatom found so far
                        metamax=posmax;
                        min=Atoms[i].cnt;
                        for (mcnt=0, j=0; j<Atoms[i].cnt; j++) // copy to array literal
                            if (Alst[Atoms[i].metalist[j]].val<0) { // unassigned atoms
                                literal[mcnt]=Atoms[i].metalist[j]; // from the metalist of i
                                if (Alst[i].pos>1)
                                    literal[mcnt]=-literal[mcnt];
                                mcnt++;
                            }
                    }
                }
            }
        }
    }
    if (req && !found)
        return 0;
    if (mcnt) { 
        for (i=0; i<mcnt; i++) { // sort atoms in array literal by decreasing precedence 
            k=i;
            max=Tmplist[abs(literal[k])];
            for (j=i+1; j<mcnt; j++) {
                if (Tmplist[abs(literal[j])]>max) {
                    max=Tmplist[abs(literal[j])];
                    k=j;
                }
            }
            if (k!=i) {
                lit=literal[i];
                literal[i]=literal[k];
                literal[k]=lit;
            }
        }
        return mcnt;
    } else if (lit) {
        literal[0]=lit;
        literal[1]=-lit;
        return 2;
    }
    return 0;
}

int lookAHead(int *literal);

// invoked by setUp()
int logic()
{
    int flag=1, more=1, i, *literal, value=1;

    literal=newData(sizeof(int)*(Nvar+1), 0);
    while (more)
        switch(flag) {
        case 1:
            flag=(LAH) ? lookAHead(literal) : getNextLit(literal);
            if (flag>0) {
                Branch++;
                for (i=1; i<flag; i++)
                    pushStck(abs(literal[i]), literal[i]>0 ? 1 : 0);
                flag=unitProp(abs(literal[0]), literal[0]>0 ? 1 : 0);
            }
            break;
        case 0:
            Backtrack++;
            if ((literal[0]=popStck(&value)))
                flag=unitProp(literal[0], value);
            else {
                if (Count && Cnt)
                    flag=-1;
                more=0;
            }
            break;
        case -1: 
            if (Answer)
                printAnswer();
            if (Count) {
                setCnt();
                if (Count>0 && Count<Cnt)
                    more=0;
                else
                    flag=1;
            } else {
                Cnt=1;
                more=0;
            }
            break;
        default :
            more=0;
            break;
        }
    return flag;
}

// invoked by preProcess() and propRules()
// returns 0 when the rule is false
int evaluateRule(int rule)
{
    int ret=1;

    if (RLst[rule].sat)
        return 1;
    if (!RLst[rule].bdycnt) {
        if (!RLst[rule].hdcnt) // empty rule
            return 0;
        else if (RLst[rule].hdcnt==1) { // only one atom, in the head
            if ((ret=propAtoms(Rules[rule].hd, Rules[rule].hdcnt, Rules[rule].horn, 1))==-1) { 
                ret=1;
                removeRef(rule);
            }
        }
    } else if (RLst[rule].bdycnt==1 && !RLst[rule].hdcnt) { // only one atom, in the body
        if ((ret=propAtoms(Rules[rule].bdy, Rules[rule].bdycnt, 0, 0))==-1) {
            ret=1;
            removeRef(rule);
        }
    }
    return ret;
}

// invoked by setUp()
// returns 0 when the theory is inconsistent, -1 when a solution is found, 1 otherwise
int preProcess()
{
    int i;

    memset(AtomQueVal, -1, sizeof(short)*(Nvar+1));
    SetAtomCnt=0;
    for (i=0; i<Nrule; i++)
        if (!evaluateRule(i)) // if the rule is empty
            return 0;
    if (SetAtomCnt && !propagate())
        return 0;
    memset(AtomQueVal, -1, sizeof(short)*(Nvar+1));
    SetAtomCnt=0;
    for (i=1; i<=Outvar; i++)
        if (Atoms[i].horn && Alst[i].req==1 && Alst[i].val<0 && AtomQueVal[i]<0) {
            if (!checkHorn(i))
                return 0;
        }
    if (SetAtomCnt && !propagate())
        return 0;
    if (Nsat<=Sat) {
        Cnt++;
        if (Answer)
            printAnswer();
        return -1;
    }
    return 1;
}

// invoked by setUp()
// assigns some precedence to rules
// the smaller the number of atoms in the rule, the larger rule precedence
void setPrec()
{
    int max=10, i;
    double prec=1;

    Prec=(double *)newData(sizeof(double)*(Maxlit+1), 0);
    for (i=Maxlit; i>max; i--)
        Prec[i]=prec;
    if (Maxlit<max)
        max=Maxlit;
    for (i=max; i>=0; i--) {
        Prec[i]=prec;
        prec *= Mult;
    }
    for (i=0; i<Nrule; i++)
        RLst[i].prec=Prec[RLst[i].bdycnt+RLst[i].hdcnt];
}

// invoked by main()
void setUp(char *filename)
{
    FILE *summary;
    static char *solution[]={"unsat", "sat", "unknown", "error"};
    int answer=0, sol=0;
    struct stat buf;

    if ((sol=getProblem())==-1)
        return;
    if (sol) {
        setReference();
        initializeCurrent();
        setPrec();
        if (Print) {
            printTheory();
            return;
        }
        answer=0;
        sol=preProcess();
        while (sol==1)
            sol=logic();
    }
    if (sol==-1)
        sol=1;
    if (stat("aspps.stat", &buf)==-1)
        answer=1;
    if ((summary=fopen("aspps.stat", "a"))==NULL) {
        fprintf(stderr, "cannot open aspps.stat");
        return;
    }
    if (answer)
        fprintf(summary, "Variable  Clauses   Sat   Branch    Bcktrck  Count     Filename\n");
    fprintf(summary, "%-10i%-10i%-6s%-10i%-10i%-10i",
            Nvar, Nrule, solution[sol], Branch, Backtrack, Cnt);
    fprintf(summary, "%s\n", filename);
    fclose(summary);
}

int main(int argc, char **argv)
{
    char filename[130];
    int i, j, fn=0;

    Print=0;
    Sat=0;
    Count=0;
    Cnt=0;
    Answer=0;
    Backtrack=0;
    ShowAtoms=NULL;
    printf("\n%s\n", VERSION);
    for (i=1; i<argc; i++)
        if (argv[i][0]=='-')
            switch(argv[i][1]) {
            case 'P':
                Print=1;
                break; 
            case 'f':
                fn=1;
                strcpy(filename, argv[++i]);
                break;
            case 'C':
                if ((i+1==argc) || (argv[i+1][0]=='-')) 
                    Count=-1;
                else
                    Count=atoi(argv[++i]);
                break;
            case 'S':
                while (ShowCnt+i+1<argc && argv[i+ShowCnt+1][0]!='-')
                    ShowCnt++;
                if ((ShowAtoms=(char **)malloc(sizeof(char *)*ShowCnt))==NULL)
                    exit(1);
                for (j=0; j<ShowCnt; )
                    ShowAtoms[j++]=strdup(argv[++i]);
                if (ShowCnt)
                    Answer=1;
                break;
            case 'L':
                LAH=((i+1==argc) || (argv[i+1][0]=='-')) ? 5 : atoi(argv[++i]);
                break;
            case 'A':
                Answer=1;
                break;
            default:
                fprintf(stderr,
                 "\nusage: aspps [-f filename] [-L [nn] ] [-C [nn]] [-A] [-S predicate] \n");
                fprintf(stderr, "-L : Adds lookahead with default of 5 atoms \n");
                fprintf(stderr, "-L nn : Adds lookahead with nn atoms \n");
                fprintf(stderr, "-C : finds all solutions.  Number of solutions written to stat file \n");
                fprintf(stderr, "-C nn: finds first nn solutions.\n");
                fprintf(stderr, "-A : Displays all positive atoms\n");
                fprintf(stderr, 
                    "-S predicate : Displays all positive atoms constructed from predicate\n");
                return 1;
            }
    if (!fn) {
        fptr=stdin;
        setUp("stdin");
    } else if ((fptr=fopen(filename, "r"))==NULL) {
        printf("error: cannot open  %s \n", filename);
        return -1;
    } else 
        setUp(filename);
    return 0;
}

int lookAHead(int *literal)
{
    int lit=0, more, pflag, i, val, scnt=0, cnt=0;

    more=LAH;
    while (more>0)
        if ((cnt=getNextLit(literal)) && cnt<3) {
            for (scnt=0, i=0; i<cnt; i++) {
                more--;
                pushStck(abs(literal[i]), literal[i]>0 ? 1 : 0);
                if ((pflag=unitProp(abs(literal[i]), literal[i]>0 ? 1 : 0))==-1)
                    return -1;
                if (pflag)
                    lit=i;
                scnt += pflag;
                popStck(&val); // value returned ignored
            }
            if (!scnt)
                return 0;
            else if (scnt==1) {
                unitProp(abs(literal[lit]), literal[lit]>0 ? 1 : 0); // value returned ignored
                more=LAH;
            }
        } else
            return cnt;
    return cnt;
}
